from __future__ import unicode_literals

import re
import unicodedata
from collections import OrderedDict
from contextlib import contextmanager
from datetime import date, time, datetime

from eztoml.errors import EzTomlEncodeError
from .tokens import (
    CONTROL_CHARS,
    SQ_INLINE,
    DQ_INLINE,
    TAB,
    UNESCAPES,
    DQ_MULTI,
    SQ_MULTI,
    NEWLINE,
)
from .utils import string_types, unicode_type, number_types

try:
    from functools import lru_cache
except ImportError:
    # ignore if unavailable
    def lru_cache(**_):
        def decorator(f):
            return f
        return decorator


@lru_cache(maxsize=1024)
def should_unicode_escape(char):
    """Determine if a character needs a unicode escape sequence."""
    if char in " \b\f\n\r\t":
        return False

    category = unicodedata.category(char)
    return category[0] in "CZ"


@lru_cache(maxsize=1024)
def escape_char(char):
    if char in UNESCAPES:
        return UNESCAPES[char]
    elif should_unicode_escape(char):
        codepoint = ord(char)
        if codepoint > 0xFFFF:
            return "\\U{:08x}".format(codepoint)

        return "\\u{:04x}".format(codepoint)
    else:
        return char


def escape(seq):
    return "".join(escape_char(c) for c in seq)


def escape_multi_line(seq):
    return "".join(escape_char(c) if c != '"' else c for c in seq)


class TokenStream(list):
    def __init__(self, nl="\n", indent=2):
        self._nl = nl
        self._prefix = ""
        self.indent = indent
        self.depth = 0
        list.__init__(self)

    def _update_prefix(self):
        self._prefix = self.depth * self.indent * " "
        return self._prefix

    @property
    def is_indented(self):  # type: () -> bool
        return self.depth > 0

    @contextmanager
    def indented(self, depth=None, delta=None):
        assert depth is not None or delta is not None
        new_depth = depth if depth is not None else delta
        depth = self.depth

        try:
            self.depth = new_depth
            self._update_prefix()
            yield self
        finally:
            self.depth = depth
            self._update_prefix()

    def __enter__(self):
        self.depth += 1
        self._update_prefix()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.depth -= 1
        self._update_prefix()

    def appendall(self, *args):
        self.extend(args)

    def flush(self):
        self.append(self._nl)

    def tab(self):
        self.append(self._prefix)

    def __str__(self):
        uc = self.__unicode__()
        if not isinstance(uc, str):
            # py2 compat
            uc = uc.encode("utf-8")
        return uc

    def __unicode__(self):
        return "".join(self)


class Table(object):
    __slots__ = (
        "is_array",
        "path",
        "kv",
        "depth",
    )

    def __init__(self, path, kv, is_array, depth=0):
        # type: (tuple[str], dict, bool, int) -> None
        self.is_array = is_array
        self.path = path
        self.kv = kv
        self.depth = depth

    def __repr__(self):
        return (
            "{self.__class__.__name__}({self.path!r}, {self.kv!r}, "
            "is_array={self.is_array!r}, depth={self.depth!r})".format(self=self)
        )


class Encoder(object):

    _key_regex = re.compile(r"^[-_A-Za-z0-9]+$")
    _word_regex = re.compile(r"[^\s]*\s*", re.MULTILINE | re.DOTALL | re.UNICODE)

    def __init__(self, sort_keys=False, nl="\n", indent=2, wrap=120, preserve_cr=False):
        self.indent = indent
        self.nl = nl
        self.sort_keys = sort_keys
        self.sorted = sorted if sort_keys else lambda x: x
        self.wrap = wrap
        self.preserve_cr = preserve_cr
        object.__init__(self)

    def encode(self, document):
        if not isinstance(document, dict):
            raise EzTomlEncodeError("Unable to encode {}".format(document))

        stream = TokenStream(indent=self.indent, nl=self.nl)
        tables = self._collect_tables(document)

        for i, table in enumerate(tables):
            if i > 0:
                stream.flush()

            self._write_table(table, stream)

        return stream.__unicode__()

    def _collect_tables(self, current, prefix=(), is_array=False):
        # type: (dict, tuple, bool) -> list[Table]
        current_keys = OrderedDict([])
        collected = []

        for k, v in self.sorted(current.items()):
            path = prefix + (k,)

            if isinstance(v, dict):
                collected.extend(self._collect_tables(v, prefix=path))
            elif isinstance(v, (list, tuple)) and len(v) > 0 and all(isinstance(vv, dict) for vv in v):
                for vv in v:
                    collected.extend(self._collect_tables(vv, prefix=path, is_array=True))
            elif v is not None:
                # toml doesn't have null, so we just skip these
                current_keys[unicode_type(k)] = v

        # only create a parent table if it can't be collapsed/inferred by the c
        if not (len(current) == 1 and isinstance(next(iter(current.values())), dict)):
            # indent all child tables unless we're at the root
            if prefix != ():
                for table in collected:
                    table.depth += 1

                collected.insert(0, Table(prefix, current_keys, is_array))
            elif len(current_keys) > 0:
                collected.insert(0, Table(prefix, current_keys, is_array))

        return collected

    def _write_table(self, table, stream):
        # type: (Table, TokenStream) -> None
        # the top most fields don't need to be indented
        with stream.indented(table.depth):
            if table.path:
                if table.is_array:
                    self._write_array_table_header(table, stream)
                else:
                    self._write_table_header(table, stream)

            if table.kv:
                self._write_kv(list(table.kv.items()), stream)

    def _write_array_table_header(self, table, stream):
        # type: (Table, TokenStream) -> None
        stream.tab()
        stream.append("[[")
        self._write_keys(table.path, stream)
        stream.append("]]")
        stream.flush()

    def _write_key(self, key, stream):  # type: (str, TokenStream) -> None
        return self._encode_key(key, stream)

    def _write_keys(self, keys, stream):  # type: (tuple, TokenStream) -> None
        for i, key in enumerate(keys):
            if i > 0:
                stream.append(".")

            self._write_key(key, stream)

    def _write_table_header(self, table, stream):
        # type: (Table, TokenStream) -> None
        stream.tab()
        stream.append("[")
        self._write_keys(table.path, stream)
        stream.append("]")
        stream.flush()

    def _write_kv(self, kv, stream):
        # type: (list[str, object], TokenStream) -> None
        for k, v in kv:
            stream.tab()
            self._encode_key(k, stream)
            stream.append(" = ")
            self._encode_value(v, stream)
            stream.flush()

    def _encode_key(self, key, stream):
        if isinstance(key, bool):
            return stream.append("true" if key else "false")

        key = unicode_type(key)

        if self._key_regex.match(key) is not None:
            stream.append(key)
        else:
            self._encode_string(key, stream, multiline=False)

    def _encode_string(self, value, stream, multiline=True):
        # type: (str, TokenStream, bool) -> None
        if isinstance(value, bytes):
            try:
                value = value.decode("utf-8")
            except UnicodeDecodeError:
                raise EzTomlEncodeError("Unable to encode bytes: {}".format(repr(value)))

        is_long = len(value) + len(stream._prefix) > 120
        multiline = multiline and ("\n" in value.lstrip() or is_long)

        if self.preserve_cr:
            multiline = multiline and "\r" not in value

        has_double_quote = DQ_INLINE in value
        has_single_quote = SQ_INLINE in value
        has_backslash = "\\" in value
        has_tab = TAB in value

        if not multiline:
            must_escape = any(k in CONTROL_CHARS or should_unicode_escape(k) for k in value)

            if must_escape is False:
                if not has_double_quote and not has_backslash:
                    # can use "..." without any escape needed
                    return self._encode_inline_string(value, stream)

                if not has_single_quote and not has_tab and (has_double_quote or has_backslash):
                    # a raw string is preferred if it has " or \ and doesn't require any other escapes
                    return self._encode_inline_raw_string(value, stream)

            return self._encode_inline_string(value, stream)
        else:
            # strip out \r because we can't make any guarantees
            # and don't want to mix file endings in the target file
            escape_chars = "".join(k for k in CONTROL_CHARS if k not in (DQ_INLINE, NEWLINE)) + TAB
            must_escape = any(k in escape_chars or should_unicode_escape(k) for k in value)

            if not must_escape and (has_backslash and SQ_MULTI not in value):
                return self._encode_multiline_raw_string(value, stream)
            else:
                return self._encode_multiline_string(value, stream)

    def _encode_multiline_raw_string(self, value, stream):  # type: (str, TokenStream) -> None
        stream.append(SQ_MULTI)
        if "\n" in value:
            stream.flush()

        stream.append(value)
        stream.append(SQ_MULTI)

    def _encode_multiline_string(self, value, stream):  # type: (str, TokenStream) -> None
        stream.append(DQ_MULTI)
        stream.flush()

        trailing_nl = value.endswith("\n")
        endswith_space = False

        for i, line in enumerate(value.splitlines()):
            if i > 0:
                # if the previous line ended with a trailing space
                # then manually print out a \n with a \ to make it clear
                # since trailing space is not typically visible
                if endswith_space:
                    stream.append("\\n\\")
                stream.flush()

            line = escape_multi_line(line).replace(DQ_MULTI, '""\\"')
            col = 0
            endswith_space = line.endswith(" ")

            for match in self._word_regex.finditer(line):
                word = match.group()

                if col == 0 or col + len(word) < self.wrap:
                    stream.append(word)
                    col += len(word)
                else:
                    stream.append("\\")
                    stream.flush()
                    stream.append(word)
                    col = len(word)

        if trailing_nl:
            stream.flush()

        stream.append(DQ_MULTI)

    def _encode_inline_raw_string(self, value, stream):  # type: (str, TokenStream) -> None
        stream.append(SQ_INLINE)
        stream.append(value)
        stream.append(SQ_INLINE)

    def _encode_inline_string(self, value, stream):  # type: (str, TokenStream) -> None
        stream.append(DQ_INLINE)
        stream.append(escape(value))
        stream.append(DQ_INLINE)

    def _encode_inline_table(self, value, stream):  # type: (str, TokenStream) -> None
        # determine if it should be split across multiple lines
        multiline = False
        running_length = 0

        for v in value:
            if isinstance(v, (list, tuple)):
                multiline = True
                break
            elif isinstance(v, string_types):
                if "\n" in v:
                    multiline = True
                    break

                running_length += len(v)
                if self.wrap is not None and running_length >= self.wrap:
                    multiline = True
                    break

        multiline = multiline and len(value) > 1
        stream.append("[")

        with stream.indented(delta=int(multiline)):
            for pos, v in enumerate(value):
                if pos != 0:
                    stream.append(",")

                    if not multiline:
                        stream.append(" ")

                if multiline:
                    stream.flush()
                    stream.tab()

                self._encode_value(v, stream)

        if multiline:
            # add a trailing comma
            stream.append(",")
            stream.flush()
            stream.tab()

        stream.append("]")

    def _encode_value(self, value, stream, multiline=True):
        if value is True:
            stream.append("true")
        elif value is False:
            stream.append("false")
        elif isinstance(value, float):
            # use repr() instead of str() because repr maintains precision better
            # for Python 2.7
            stream.append(repr(value))
        elif isinstance(value, number_types):
            stream.append(str(value))
        elif isinstance(value, string_types):
            self._encode_string(value, stream, multiline=multiline)
        elif isinstance(value, dict):
            stream.append("{  ")
            for pos, (k, v) in enumerate(self.sorted(value.items())):
                if pos != 0:
                    stream.append(", ")

                self._encode_key(k, stream)
                stream.append(" = ")
                # don't allow multi-line strings within inline tables
                self._encode_value(v, stream)

            stream.append("}")
        elif isinstance(value, list):
            self._encode_inline_table(value, stream)
        elif isinstance(value, datetime):
            stream.append(value.isoformat())
        elif isinstance(value, time):
            stream.append(str(value))
        elif isinstance(value, date):
            stream.append(str(value))
        else:
            raise EzTomlEncodeError("Unable to encode {}".format(value))



"""Easy TOML."""
from __future__ import print_function

from eztoml.encoder import Encoder
from .decoder import Decoder
from .errors import EzTomlDecodeError, EzTomlEncodeError, EzTomlError
from .tz import EzTomlTz

__version__ = "0.0.1.dev1"


def loads(src, **kwargs):
    return Decoder(**kwargs).decode(src)


def load(f, **kwargs):
    return loads(f.read(), **kwargs)


def dumps(document, **kwargs):
    if not isinstance(document, dict):
        raise EzTomlEncodeError("Unable to encode non-dictionary type: {}".format(type(src)))

    return Encoder(**kwargs).encode(document)


def dump(document, fileobj, **kwargs):
    fileobj.write(dumps(document, **kwargs))


def lint_files(args=None):
    import io
    import fnmatch
    import glob
    import os
    import sys

    from argparse import ArgumentParser

    parser = ArgumentParser("toml-lint")
    parser.add_argument("paths", metavar="file_or-directory", nargs="+")
    parser.add_argument("-r", "--recursive", action="store_true")
    parsed = parser.parse_args(args)

    matches = []

    for path in parsed.paths:
        if os.path.isfile(path):
            matches.append(path)
        elif os.path.isdir(path):
            if parsed.recursive:
                for root, _, files in os.walk(path):
                    for filename in fnmatch.filter(files, '*.toml'):
                        matches.append(os.path.join(root, filename))
            else:
                matches.extend(glob.glob(os.path.join(path, "*.toml")))

    for path in matches:
        print(path)

        try:
            with io.open(path, "rt", encoding="utf-8") as infile:
                decoded = load(infile)

            # load to an intermediate variable, so that we don't wipe out the file
            # if there's an error after the handle is opened
            encoded = dumps(decoded)

            with io.open(path, "wt", encoding="utf-8") as outfile:
                outfile.write(encoded)

        except EzTomlError as exc:
            print(exc, file=sys.stderr)



from __future__ import unicode_literals

import datetime
import re

from .errors import EzTomlDecodeError
from .source import Source
from .tokens import (
    DQ_MULTI,
    DQ_INLINE,
    SQ_MULTI,
    SQ_INLINE,
    NAN,
    POS_INF,
    NEG_INF,
    RE_FLAGS,
    CONTROL_CHARS,
    ESCAPES,
)
from .tz import EzTomlTz
from .utils import string_types

try:
    from_codepoint = unichr
except NameError:
    from_codepoint = chr


class Decoder(object):

    __escapes = ESCAPES
    _time_regex = re.compile(r"^(\d{2}):(\d{2}):(\d{2})(?:\.(\d{3,}))?")
    _date_regex = re.compile(r"^(\d{4})-(\d{2})-(\d{2})")
    _tz_regex = re.compile(r"^(Z|[-+]\d{2}:\d{2})")
    _datetime_regex = re.compile(
        _date_regex.pattern[1:] + r"(?:[T ]" + _time_regex.pattern[1:] + ")" + _tz_regex.pattern[1:] + "?"
    )
    _hex_regex = re.compile(r"^0x[A-Za-z0-9](?:_?[A-Za-z0-9])*")
    _octal_regex = re.compile(r"^0o[0-7](?:_?[0-7])*")
    _int_regex = re.compile(r"^[-+]?[0-9](?:_?[0-9])*")
    _binary_regex = re.compile(r"^0b[0-1](?:_?[0-1])*")
    _float_regex = re.compile(r"^[-+]?[0-9](?:_?[0-9])*(?:\.[0-9](?:_?[0-9])*)?(?:[eE][+-]?[0-9](?:_?[0-9])*)?")
    _key_regex = re.compile(r"^[-_A-Za-z0-9]+", RE_FLAGS)
    _is_hex4 = staticmethod(re.compile(r"^[A-Za-z0-9]{4}").match)
    _is_hex8 = staticmethod(re.compile(r"^[A-Za-z0-9]{8}").match)
    _get_escape = staticmethod(__escapes.get)
    _is_control_char = staticmethod(CONTROL_CHARS.__contains__)

    def decode(self, source):
        if isinstance(source, bytes):
            source = source.decode("utf-8")

        if isinstance(source, string_types):
            source = Source(source)
        elif not isinstance(source, Source):
            raise EzTomlDecodeError("Expected a Source or String to decode")

        source.eat_ws()
        document = self._decode_root(source)
        source.eat_ws()

        if not source.eof:
            raise EzTomlDecodeError("Extraneous input")

        return document

    def _decode_root(self, source):
        document = {}
        source.eat_ws()

        while not source.eof:
            if source.has_prefix("[["):
                self._decode_table_array(source, document)
            elif source.has_prefix("["):
                self._decode_table(source, document, allow_existing=True)
            elif source.peek(1) in (SQ_INLINE, DQ_INLINE) or source.peek_match(self._key_regex):
                self._decode_kv(source, document)
            else:
                raise EzTomlDecodeError("Unknown input")

            source.eat_ws()

        return document

    def _make_table_path(self, path, document, check_arrays=True):
        sub_table = document
        for k in path:
            if k not in sub_table:
                sub_table = sub_table.setdefault(k, {})
                continue

            sub_table = sub_table[k]

            if not isinstance(sub_table, dict):
                if check_arrays and isinstance(sub_table, list) and len(sub_table) > 0:
                    if len(sub_table) > 0 and all(isinstance(t, dict) for t in sub_table):
                        sub_table = sub_table[-1]
                        continue

                raise EzTomlDecodeError("Key path overlaps with existing value")

        return sub_table

    def _decode_kv(self, source, table):
        keys = set()

        while not source.eof and not source.has_prefix("["):
            key = self._decode_key(source)

            # if no key was found, then we're done
            if key == ():
                raise EzTomlDecodeError("Expected a key")

            # can't redefine a key within an already constructed dict/table
            for subkey_len in range(1, len(key)):
                subkey = key[:subkey_len]
                if subkey in keys:
                    raise EzTomlDecodeError(
                        "Can't add key {key} when {subkey} is already defined".format(key=key, subkey=subkey)
                    )

            keys.add(key)

            source.eat_inline_ws()
            if source.take(1) != "=":
                raise EzTomlDecodeError("Missing = after table key")

            source.eat_inline_ws()
            sub_table = self._make_table_path(key[:-1], table)

            if key[-1] in sub_table:
                raise EzTomlDecodeError("Duplicate key {}".format(key))

            value = self._decode_value(source)

            source.eat_ws(must_advance=True)
            sub_table[key[-1]] = value
            keys.add(key)

        return table

    def _decode_table(self, source, document, allow_existing=False):
        if not source.has_prefix("[") or source.has_prefix("[["):
            raise EzTomlDecodeError("Expected table")

        source.take(1)
        source.eat_inline_ws()
        path = self._decode_key(source)
        parent_table = self._make_table_path(path[:-1], document)

        if path[-1] in parent_table:
            existing = parent_table[path[-1]]
            if not isinstance(existing, dict):
                raise EzTomlDecodeError("Value already defined as {}".format(type(existing).__name__.lower()))

            has_kv = any(not isinstance(v, dict) for v in existing.values())
            if allow_existing is False or has_kv:
                raise EzTomlDecodeError("Duplicated table")

        # create a table if it doesn't exist, otherwise take the one that does
        table = parent_table.setdefault(path[-1], {})
        source.eat_inline_ws()

        if source.take(1) != "]":
            raise EzTomlDecodeError("Unclosed table initializer. Expected: ]")

        source.eat_ws(must_advance=True)
        return self._decode_kv(source, table)

    def _decode_table_array(self, source, document):
        if source.take(2) != "[[":
            raise EzTomlDecodeError("Expected table")

        source.eat_inline_ws()
        path = self._decode_key(source)
        parent_table = self._make_table_path(path[:-1], document)
        table = {}

        if isinstance(parent_table.get(path[-1]), list) and len(parent_table[path[-1]]) == 0:
            raise EzTomlDecodeError("Can't add table to existing list")

        if not isinstance(parent_table.setdefault(path[-1], []), list):
            raise EzTomlDecodeError("Duplicated table")

        parent_table[path[-1]].append(table)
        source.eat_inline_ws()

        if source.take(2) != "]]":
            raise EzTomlDecodeError("Unclosed table array initializer. Expected: ]]")

        source.eat_ws(must_advance=True)
        return self._decode_kv(source, table)

    def _decode_key(self, source):  # type: (Source) -> tuple[str]
        path = []

        while not source.eof:
            source.eat_inline_ws()
            char = source.peek(1)
            dotted = char == "."

            if dotted:
                if not path:
                    raise EzTomlDecodeError("Unexpected dotted key")

                source.take(1)
                source.eat_inline_ws()
                char = source.peek(1)

            key_peek = source.peek_match(self._key_regex)

            if key_peek is not None:
                path.append(source.take(len(key_peek)))
            elif char == DQ_INLINE:
                path.append(self._decode_escaped_str(source))
            elif char == SQ_INLINE:
                path.append(self._decode_literal_str(source))
            else:
                if dotted and not path:
                    raise EzTomlDecodeError("Unmatched dot for key")

                return tuple(path)

        raise EzTomlDecodeError("Unexpected EOF while parsing key")

    def _decode_inline_table(self, source):
        if not source.remove_prefix("{"):
            raise EzTomlDecodeError("Expected {")

        source.eat_ws()
        table = {}

        while not source.eof:
            if source.remove_prefix("}"):
                return table

            if table:
                if not source.remove_prefix(","):
                    raise EzTomlDecodeError("Expected , or }")
                source.eat_inline_ws()

            key = self._decode_key(source)
            parent_table = self._make_table_path(key[:-1], table, check_arrays=False)
            if key[-1] in parent_table:
                raise EzTomlDecodeError("Duplicate key for inline table")

            source.eat_inline_ws()
            if not source.remove_prefix("="):
                raise EzTomlDecodeError("Missing = for inline table key")

            source.eat_inline_ws()
            parent_table[key[-1]] = self._decode_value(source)
            source.eat_inline_ws()

        raise EzTomlDecodeError("Expected } not EOF")

    def _decode_inline_array(self, source):
        if source.take(1) != "[":
            raise EzTomlDecodeError("Expected [")

        source.eat_ws()
        array = []

        while not source.eof:
            if array:
                if source.remove_prefix(","):
                    source.eat_ws()
                elif not source.has_prefix("]"):
                    break

            if source.remove_prefix("]"):
                return array

            array.append(self._decode_value(source))
            source.eat_ws()

        raise EzTomlDecodeError("Expected ]")

    def _decode_str(self, source):
        if source.has_prefix(DQ_MULTI):
            return self._decode_escaped_str_multiline(source)
        elif source.has_prefix(SQ_MULTI):
            return self._decode_literal_str_multiline(source)
        elif source.has_prefix(DQ_INLINE):
            return self._decode_escaped_str(source)
        elif source.has_prefix(SQ_INLINE):
            return self._decode_literal_str(source)
        else:
            raise EzTomlDecodeError("Unknown string type")

    @classmethod
    def _unescape(cls, source):
        # \uXXXX     - unicode         (U+XXXX)
        # \UXXXXXXXX - unicode         (U+XXXXXXXX)
        char = source.take(1)
        if char in "uU":
            size = 4 if char == "u" else 8
            match = cls._is_hex4 if char == "u" else cls._is_hex8
            scalar = source.take(size)
            if match(scalar) is None:
                raise EzTomlDecodeError("Expected a valid unicode scalar value")

            try:
                return from_codepoint(int(scalar, 16))
            except ValueError:
                raise EzTomlDecodeError("Invalid unicode scalar value")

        else:
            unescaped = cls._get_escape(char)
            if unescaped is None:
                raise EzTomlDecodeError("Unknown escape sequence")
            return unescaped

    def _decode_escaped_str_multiline(self, source):  # type: (Source) -> str
        # Basic strings are surrounded by quotation marks.
        if source.take(3) != DQ_MULTI:
            raise EzTomlDecodeError("Expected {}".format(DQ_MULTI))

        source.remove_prefix("\n") or source.remove_prefix("\r\n")
        pieces = []

        while not source.eof:
            if source.has_prefix(DQ_MULTI):
                source.take(3)

                # strings ending with """" -> "
                if source.has_prefix(DQ_INLINE):
                    pieces.append(source.take(1))

                # strings ending with """"" -> ""
                if source.has_prefix(DQ_INLINE):
                    pieces.append(source.take(1))

                return "".join(pieces)

            char = source.take(1)

            if char == "\\":
                if source.has_prefix("\n") or source.peek(2) == "\r\n":
                    # When the last non-whitespace character on a line is an unescaped \,
                    # it will be trimmed along with all whitespace (including newlines)
                    # up to the next non-whitespace character or closing delimiter.
                    source.eat_ws()
                    continue

                pieces.append(self._unescape(source))
            elif char not in "\r\t\n" and self._is_control_char(char):
                raise EzTomlDecodeError("Invalid use of unescaped control character")
            else:
                pieces.append(char)

        raise EzTomlDecodeError("Unexpected EOF while waiting for {}".format(DQ_MULTI))

    def _decode_literal_str_multiline(self, source):
        if source.take(3) != SQ_MULTI:
            raise EzTomlDecodeError("Expected {}".format(SQ_MULTI))

        source.remove_prefix("\n") or source.remove_prefix("\r\n")
        pieces = []

        while not source.eof:
            if source.has_prefix(SQ_MULTI):
                source.take(3)

                # strings ending with '''' -> '
                if source.peek(1) == SQ_INLINE:
                    pieces.append(source.take(1))

                # strings ending with ''''' -> ''
                if source.peek(1) == SQ_INLINE:
                    pieces.append(source.take(1))

                return "".join(pieces)

            char = source.take(1)

            if char not in "\r\t\n" and self._is_control_char(char):
                raise EzTomlDecodeError("Invalid use of unescaped control character")
            else:
                pieces.append(char)

        raise EzTomlDecodeError("Unexpected EOF while waiting for {}".format(SQ_MULTI))

    def _decode_escaped_str(self, source):
        if source.take(1) != DQ_INLINE:
            raise EzTomlDecodeError("Expected {}".format(DQ_INLINE))

        pieces = []

        while not source.eof:
            if source.peek(1) == DQ_INLINE:
                source.take(1)
                return "".join(pieces)

            char = source.take(1)

            if char in "\r\n":
                raise EzTomlDecodeError("Unexpected EOL while parsing newline character")
            elif char == "\\":
                pieces.append(self._unescape(source))
            elif self._is_control_char(char):
                raise EzTomlDecodeError("Invalid use of unescaped control character")
            else:
                pieces.append(char)

        raise EzTomlDecodeError("Unexpected EOF while waiting for {}".format(DQ_INLINE))

    def _decode_literal_str(self, source):
        # Basic strings are surrounded by quotation marks.
        if source.take(1) != SQ_INLINE:
            raise EzTomlDecodeError("Expected {}".format(SQ_INLINE))

        pieces = []

        while not source.eof:
            if source.has_prefix(SQ_INLINE):
                source.take(1)
                return "".join(pieces)

            char = source.take(1)

            if char == "'":
                return "".join(pieces)
            elif char in "\r\n":
                raise EzTomlDecodeError("Unexpected EOL while parsing newline character")
            elif char != "\t" and self._is_control_char(char):
                raise EzTomlDecodeError("Invalid use of unescaped control character")
            else:
                pieces.append(char)

        raise EzTomlDecodeError("Unexpected EOF while waiting for {}".format(SQ_INLINE))

    @staticmethod
    def _to_int(src, base):
        return int(src.replace("_", ""), base)

    def _decode_number(self, source):  # type: (Source) -> int|float
        try:
            if source.has_prefix("0x"):
                return self._to_int(source.take_match(self._hex_regex), 16)
            elif source.has_prefix("0o"):
                return self._to_int(source.take_match(self._octal_regex), 8)
            elif source.has_prefix("0b"):
                return self._to_int(source.take_match(self._binary_regex), 2)
            else:
                int_match = source.peek_match(self._int_regex)
                float_match = source.peek_match(self._float_regex)

                assert int_match is not None

                # check for duplicate leading zeros
                leading = int_match.lstrip("-+").split(".")[0]
                if leading != "0" and leading.startswith("0"):
                    raise EzTomlDecodeError("Invalid leading zeros")

                assert float_match is not None and int_match is not None

                if float_match == int_match:
                    return self._to_int(source.take(len(int_match)), 10)
                else:
                    return float(source.take(len(float_match)).replace("_", ""))

        except AssertionError as exc:
            raise EzTomlDecodeError("Did not find a valid number")

    def _decode_value(self, source):
        check_time = source.peek_match(self._time_regex)
        check_date = source.peek_match(self._date_regex)
        check_datetime = source.peek_match(self._datetime_regex)

        if check_time:
            source.take(len(check_time))
            groups = list(filter(None, self._time_regex.match(check_time).groups()))

            # pad and truncate to microsecond precision
            if len(groups) == 4:
                groups[-1] = groups[-1].ljust(6, "0")[:6]

            try:
                return datetime.time(*[int(g) for g in groups])
            except ValueError:
                raise EzTomlDecodeError("Invalid RFC-3399 time")
        elif check_datetime:
            source.take(len(check_datetime))
            groups = list(self._datetime_regex.match(check_datetime).groups())
            tz_string = groups.pop()
            tz_info = EzTomlTz(tz_string) if tz_string else None

            groups = list(filter(None, groups))

            # pad and truncate to microsecond precision
            if len(groups) == 7:
                groups[-1] = groups[-1].ljust(6, "0")[:6]

            try:
                return datetime.datetime(*[int(g) for g in groups], tzinfo=tz_info)
            except ValueError:
                raise EzTomlDecodeError("Invalid RFC-3399 time")
        elif check_date:
            source.take(len(check_date))
            groups = list(self._date_regex.match(check_date).groups())

            try:
                return datetime.date(*[int(g) for g in groups])
            except ValueError:
                raise EzTomlDecodeError("Invalid RFC-3399 time")
        elif source.has_prefix("{"):
            return self._decode_inline_table(source)
        elif source.has_prefix("["):
            return self._decode_inline_array(source)
        elif source.peek(1) in "'\"":
            return self._decode_str(source)
        elif source.remove_prefix("true"):
            return True
        elif source.remove_prefix("false"):
            return False
        elif source.remove_prefix("+nan") or source.remove_prefix("-nan") or source.remove_prefix("nan"):
            return NAN
        elif source.remove_prefix("+inf") or source.remove_prefix("inf"):
            return POS_INF
        elif source.remove_prefix("-inf"):
            return NEG_INF
        elif source.peek(1) in "+-" or source.peek(1).isdigit():
            return self._decode_number(source)
        else:
            raise EzTomlDecodeError("Missing value")



from __future__ import unicode_literals

import re  # noqa: F401

from .errors import EzTomlDecodeError


class Source(object):
    def __init__(self, text):  # type: (str) -> None
        self.text = text
        self.size = len(self.text)
        self.pos = 0
        self.line = 0
        self.column = 0

    def copy(self):
        copied = type(self)(self.text)
        copied.pos = self.pos
        copied.line = self.line
        copied.column = self.column

    def reset(self):
        self.pos = 0
        self.line = 0
        self.column = 0

    @property
    def eof(self):
        return self.size == self.pos

    @property
    def remaining(self):
        return self.text[self.pos :]

    @property
    def remaining_line(self):
        remaining, _, _ = self.remaining.partition("\n")
        return remaining.strip()

    def peek(self, num_chars=1):
        if self.pos < self.size:
            return self.text[self.pos : self.pos + num_chars]

    def peek_match(self, regex):  # type: (re.Pattern) -> str
        matched = regex.match(self.remaining)
        if matched is not None:
            return self.text[self.pos + matched.start() : self.pos + matched.end()]

    def take_match(self, regex):  # type: (re.Pattern) -> str
        matched = regex.match(self.remaining)
        assert matched is not None
        return self.take(matched.end() - matched.start())

    def advance_line(self):
        try:
            self.pos = self.text.index("\n", self.pos) + 1
            self.line += 1
            self.column = 0
        except ValueError:
            self.column += self.size - self.pos
            self.pos = self.size

    def search(self, substring):
        try:
            return self.text.index(substring, self.pos) - self.pos
        except ValueError:
            return None

    def has_prefix(self, prefix):
        return self.peek(len(prefix)) == prefix

    def remove_prefix(self, prefix):
        if self.text[self.pos : self.pos + len(prefix)] == prefix:
            self.take(len(prefix))
            return True
        return False

    def take(self, num_chars=1):
        taken = self.text[self.pos : self.pos + num_chars]
        size = len(taken)
        assert size == num_chars
        self.pos += size
        num_lines = taken.count("\n")
        if num_lines == 0:
            self.column += size
        else:
            self.line += num_lines
            self.column = (size - 1) - taken.rindex("\n")

        return taken

    def eat_inline_ws(self):
        while not self.eof and self.text[self.pos] in " \t":
            self.pos += 1
            self.column += 1

    def eat_ws(self, must_advance=False):
        start_line = self.line

        while not self.eof:
            char = self.text[self.pos]
            if char == "#":
                self.advance_line()
                continue
            elif char == "\n":
                self.line += 1
                self.column = 0
            elif self.text[self.pos] in " \t\r":
                self.column += 1
            else:
                break

            self.pos += 1

        # if the line hasn't advanced, then we have too much content on a single line
        if must_advance and self.line == start_line and not self.eof:
            raise EzTomlDecodeError("Unexpected content in line: {}".format(self.remaining_line))



"""EzToml time zone handling."""
from datetime import tzinfo, timedelta


class EzTomlTz(tzinfo):
    """RFC-3339 compatible timezone information for TOML."""

    def __init__(self, offset_str):  # type: (str) -> None
        if offset_str == "Z":
            self.delta = timedelta()
        else:
            sign = -1 if offset_str[0] == "-" else 1
            h, m = offset_str[1:].split(":")
            self.delta = sign * timedelta(hours=int(h), minutes=int(m))

    def utcoffset(self, dt):
        """Get the offset from UTC."""
        return self.delta

    def dst(self, dt):
        """Return nothing since RFC-3339 doesn't accept local timezones."""
        return timedelta(0)

    def tzname(self, dt):
        """Convert the timezone to a time string."""
        if self.delta == timedelta():
            return "Z"

        sign = "-" if self.delta < timedelta() else "+"
        delta = abs(self.delta)
        hours = delta.total_seconds() // 60
        minutes = delta.total_seconds() % 60
        return "UTC{sign}{hours:02d}:{minutes:02d}".format(sign=sign, hours=hours, minutes=minutes)



from __future__ import unicode_literals
import re

NEWLINE = "\n"
DQ_MULTI = '"""'
DQ_INLINE = '"'
SQ_MULTI = "'''"
SQ_INLINE = "'"
NAN = float("NaN")
POS_INF = float("+inf")
NEG_INF = float("-inf")
RE_FLAGS = re.DOTALL | re.UNICODE | re.MULTILINE

# from https://github.com/toml-lang/toml
#
# \b         - backspace       (U+0008)
# \t         - tab             (U+0009)
# \n         - linefeed        (U+000A)
# \f         - form feed       (U+000C)
# \r         - carriage return (U+000D)
# \"         - quote           (U+0022)
# \\         - backslash       (U+005C)
# Any Unicode character may be used except those that must be escaped:
# - quotation mark
# - backslash
# - and the control characters other than tab (U+0000 to U+0008, U+000A to U+001F, U+007F).
CONTROL_CHARS = "".join([chr(u) for u in range(0x0, 0x8 + 1)] + [chr(u) for u in range(0xA, 0x1F)] + [chr(0x7F)])

TAB = "\t"
ESCAPES = {
    "b": "\b",
    "t": TAB,
    "n": NEWLINE,
    "f": "\f",
    "r": "\r",
    '"': '"',
    "\\": "\\",
}
UNESCAPES = {v: "\\" + k for k, v in ESCAPES.items()}



#! /usr/bin/env python

"""Perform setup of the package for build."""
import re
import io

from setuptools import setup


with io.open('eztoml/__init__.py', 'rt', encoding='utf8') as f:
    __version__ = re.search(r'__version__ = "(.*?)"', f.read()).group(1)


setup(
    name='eztoml',
    version=__version__,
    description='Easy TOML',
    author='Ross Wolf',
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=['eztoml'],
    tests_require=["PyYAML~=5.3"],
    entry_points={
        'console_scripts': [
            'toml-lint=eztoml:lint_files',
        ],
        'pygments.lexers': [
            'eql=eql.highlighters:EqlLexer'
        ]
    },
)
